from h1st.model.ml_model import MLModel
from h1st.model.ml_modeler import MLModeler


class Student(MLModel):
    """
    Knowledge Generalization Model
    """
    pass


class StudentModeler(MLModeler):
    """
    Knowledge Generalization Modeler
    """
    pass
